"""
Configuration classes definition.

Author:
    Andrea Cervesato <andrea.cervesato@mailbox.org>
"""
import logging
import configparser
import json
import yaml
import flatten_dict


class Config:
    """
    Base configuration to implement in order to push/pull configurations inside
    an etcd database.
    """

    def __init__(self, client, basefolder="/config"):
        """
        Args:
            client (etcd3.Client): etcd Client instance.
        """
        self._logger = logging.getLogger("converter")
        self._client = client
        self._basefolder = basefolder

    def _convert(self, filepath):
        """
        Convert a file into a dictionary.

        Args:
            filepath (str): path of the file to be pushed.
        """
        raise NotImplementedError()

    def push(self, name, filepath):
        """
        Push a format supported file into an etcd database.

        Args:
            name     (str): name to associate with file.
            filepath (str): path of the file to be pushed.
        """
        if not name or not isinstance(name, str):
            raise ValueError("name must be a string")

        if not filepath or not isinstance(filepath, str):
            raise ValueError("filepath must be a string")

        self._logger.info("pushing '%s' with name '%s'", filepath, name)

        config_path = "{0}/{1}".format(self._basefolder, name)

        # convert  to dict
        data = self._convert(filepath)

        # flatten the dictionary
        def slash_reducer(k1, k2):
            if k1 is None:
                return k2

            return "{0}/{1}".format(k1, k2)

        paths = flatten_dict.flatten(data, reducer=slash_reducer)

        for dirs, value in paths.items():
            path = "{0}/{1}".format(config_path, dirs)
            self._logger.debug("setting: %s -> %s", path, value)
            self._client.put(path, value)

        self._logger.info("configuration pushed")

    def pull(self, name):
        """
        Pull a format supported configuration from an etcd database.

        Args:
            name (str): name to associate with file.

        Returns:
            dict: configuration stored inside the database.
        """
        if not name or not isinstance(name, str):
            raise ValueError("name must be a string")

        self._logger.info("fetching '%s'", name)

        config_path = "{0}/{1}".format(self._basefolder, name)
        keys = self._client.get_all()

        flat_dict = dict()
        for value, metadata in keys:
            key = metadata.key.decode('utf-8')
            if key.startswith(config_path):
                key_done = key.replace(config_path, "", 1)
                flat_dict[key_done] = value.decode('utf-8')

        if not flat_dict:
            return dict()

        self._logger.info("config_path = %s", config_path)
        self._logger.info("flat_dict = %s", flat_dict)

        def slash_reducer(flat_key):
            # first element is empty
            return flat_key.split("/")[1:]

        config = flatten_dict.unflatten(flat_dict, splitter=slash_reducer)

        self._logger.info("configuration fetched")

        return config


class JsonConfig(Config):
    """
    Push/pull JSON configurations inside an etcd database.
    """

    def _convert(self, filepath):
        data = dict()
        with open(filepath, 'r') as fdata:
            data = json.loads(fdata.read())
        return data


class YamlConfig(Config):
    """
    Push/pull Yaml configurations inside an etcd database.
    """

    def _convert(self, filepath):
        data = dict()
        with open(filepath, 'r') as fdata:
            data = yaml.safe_load(fdata.read())
        return data


class IniConfig(Config):
    """
    Push/pull Yaml configurations inside an etcd database.
    """

    def _convert(self, filepath):
        parser = configparser.ConfigParser()
        parser.read(filepath)
        data = {section: dict(parser.items(section))
                for section in parser.sections()}
        return data



"""
package definition.

Author:
    Andrea Cervesato <andrea.cervesato@mailbox.org>
"""
import etcd3
import etcdgo.config


def get_config(client, config_type, basefolder="/config"):
    """
    Return an object that can be used to push/pull configurations inside
    an etcd database.

    Examples:

        import etcd3
        import etcdgo

        client = etcd3.Etcd3Client()

        # push a json configuration inside database
        config = etcdgo.get_config(client, "json")
        config.push("myconfig", "myfile.json")

        # push a yaml configuration inside database
        config = etcdgo.get_config(client, "yaml")
        config.push("myconfig", "myfile.yaml")

        # pull data from etcd database
        data = config.pull("myconfig")

    Args:
        client (etcd3.Etcd3Client): etcd client object.
        config_type    (str): configuration type. Supported: json, yaml.
        basefolder     (str): root of the configuration inside the etcd database.

    Returns:
        Config: object to push/pull configurations inside an etcd database.
    """
    if not client or not isinstance(client, etcd3.Etcd3Client):
        raise ValueError("client must be of type etcd3.Etcd3Client")

    if not config_type or not isinstance(config_type, str):
        raise ValueError("config_type must be a string")

    obj = None
    if config_type.lower() == "json":
        obj = etcdgo.config.JsonConfig(client, basefolder=basefolder)
    elif config_type.lower() == "yaml":
        obj = etcdgo.config.YamlConfig(client, basefolder=basefolder)
    elif config_type.lower() == "ini":
        obj = etcdgo.config.IniConfig(client, basefolder=basefolder)
    else:
        raise NotImplementedError("'%s' format is not supported" % config_type)

    return obj



# -*- coding: utf-8 -*-
"""
setuptools script.

Author:
    Andrea Cervesato <andrea.cervesato@mailbox.org>
"""
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='etcdgo',
    version='1.1.0',
    description='A library to push/pull configurations inside etcd databases',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Andrea Cervesato',
    author_email='andrea.cervesato@mailbox.org',
    url='https://github.com/acerv/etcdgo',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Testing',
        'Topic :: Utilities',
    ],
    packages=["etcdgo"],
    python_version=">3.5,<3.9",
    install_requires=[
        'etcd3 <= 0.12.0',
        'pyyaml <= 5.3.1',
        'flatten-dict <= 0.2.0'
    ],
)
